IMPLEMENTATION MODULE PL0GEN;

FROM SYSTEM IMPORT ADR;
IMPORT Lib, FIO;
FROM PL0INT IMPORT TFunction, TLevel, TAddress, TOperation, TValue;
FROM PL0INT IMPORT code, out, dbg;
FROM PL0INT IMPORT WrIns;
IMPORT PL0INT;

CONST
  debug = FALSE;

TYPE
  TMnemStr = ARRAY [ 0 .. 2 ] OF CHAR;

VAR
  lab: TAddress;  (* current label *)
  mnm: ARRAY TFunction OF TMnemStr;

PROCEDURE WrFix( a: TAddress );
BEGIN
  FIO.WrStr( dbg, 'fixup at ' );
  FIO.WrCard( dbg, a, 0 );
  FIO.WrLn( dbg );
END WrFix;


PROCEDURE Init();
BEGIN
  FIO.WrStr( out, '# Generating...' );
  FIO.WrLn( out );
  Lib.Fill( ADR(code), SIZE(code), fStp );
  lab := MIN( TAddress );
END Init;

PROCEDURE Label(): TAddress;
BEGIN
  RETURN lab;
END Label;

PROCEDURE Gen( f, l: CARDINAL; v: INTEGER );
BEGIN
  WITH code[ lab ] DO
    fct := VAL( TFunction, f );
    lvl := l;
    val := v;
  END;
  IF debug THEN WrIns( dbg, lab ) END;
  IF lab >= MAX( TAddress ) THEN HALT END;
  INC( lab );
END Gen;

PROCEDURE Fixup( a: TAddress );
BEGIN
  IF debug THEN WrFix( a ) END;
  code[ a ].adr := lab;
END Fixup;

PROCEDURE End();
BEGIN
  FIO.WrStr( out, '# Generation of ' );
  FIO.WrCard( out, lab, 0 );
  FIO.WrStr( out, ' instructions done.' );
  FIO.WrLn( out );
END End;

(* convenient add-ons *)

PROCEDURE List();
BEGIN
  PL0INT.List( out );
END List;

(* more convenient Gen procs *)

PROCEDURE Lit( v: TValue );
BEGIN
  Gen( ORD(fLit), 0, INTEGER(v) );
END Lit;

PROCEDURE Opr( o: TOperation );
BEGIN
  Gen( ORD(fOpr), 0, INTEGER(ORD(o)) );
END Opr;

PROCEDURE Lod( l: TLevel; a: TAddress );
BEGIN
  Gen( ORD(fLod), l, INTEGER(a) );
END Lod;

PROCEDURE Sto( l: TLevel; a: TAddress );
BEGIN
  Gen( ORD(fSto), l, INTEGER(a) );
END Sto;

PROCEDURE Cal( l: TLevel; a: TAddress );
BEGIN
  Gen( ORD(fCal), l, INTEGER(a) );
END Cal;

PROCEDURE Int( da: TAddress );
BEGIN
  Gen( ORD(fInt), 0, INTEGER(da) );
END Int;

PROCEDURE Jmp( a: TAddress );
BEGIN
  Gen( ORD(fJmp), 0, INTEGER(a) );
END Jmp;

PROCEDURE Jpc( a: TAddress );
BEGIN
  Gen( ORD(fJpc), 0, INTEGER(a) );
END Jpc;

PROCEDURE Stp();
BEGIN
  Gen( ORD(fStp), 0, 0 );
END Stp;


(* most convenient Gen procs *)

PROCEDURE Put( v: TValue );
BEGIN
  Lit( v );
END Put;

PROCEDURE Nop();
BEGIN
  Opr( opNop );
END Nop;

PROCEDURE Neg();
BEGIN
  Opr( opNeg );
END Neg;

PROCEDURE Odd();
BEGIN
  Opr( opOdd );
END Odd;

PROCEDURE Read();
BEGIN
  Opr( opRed );
END Read;

PROCEDURE Write();
BEGIN
  Opr( opWrt );
END Write;

PROCEDURE Add();
BEGIN
  Opr( opAdd );
END Add;

PROCEDURE Sub();
BEGIN
  Opr( opSub );
END Sub;

PROCEDURE Mul();
BEGIN
  Opr( opMul );
END Mul;

PROCEDURE Div();
BEGIN
  Opr( opDiv );
END Div;

PROCEDURE Eql();
BEGIN
  Opr( opEql );
END Eql;

PROCEDURE Neq();
BEGIN
  Opr( opNeq );
END Neq;

PROCEDURE Lss();
BEGIN
  Opr( opLss );
END Lss;

PROCEDURE Leq();
BEGIN
  Opr( opLeq );
END Leq;

PROCEDURE Gtr();
BEGIN
  Opr( opGtr );
END Gtr;

PROCEDURE Geq();
BEGIN
  Opr( opGeq );
END Geq;

PROCEDURE Load( l: TLevel; a: TAddress );
BEGIN
  Lod( l, a );
END Load;

PROCEDURE Store( l: TLevel; a: TAddress );
BEGIN
  Sto( l, a );
END Store;

PROCEDURE Enter( da: TAddress );
BEGIN
  Int( da );
END Enter;

PROCEDURE Leave();
BEGIN
  Opr( opRet );
END Leave;

PROCEDURE JmpZ( a: TAddress );
BEGIN
  Jpc( a );
END JmpZ;

PROCEDURE JmpFwd( VAR a: TAddress );
BEGIN
  a := Label();
  Jmp( 0 );
END JmpFwd;

PROCEDURE JmpZFwd( VAR a: TAddress );
BEGIN
  a := Label();
  JmpZ( 0 );
END JmpZFwd;

PROCEDURE Call( l: TLevel; a: TAddress );
BEGIN
  Cal( l, a );
END Call;

PROCEDURE Stop();
BEGIN
  Stp();
END Stop;

BEGIN
  mnm[ fLit ] := "LIT";
  mnm[ fOpr ] := "OPR";
  mnm[ fLod ] := "LOD";
  mnm[ fSto ] := "STO";
  mnm[ fCal ] := "CAL";
  mnm[ fInt ] := "INT";
  mnm[ fJmp ] := "JMP";
  mnm[ fJpc ] := "JPC";
END PL0GEN.
