IMPLEMENTATION MODULE PL0INT;

IMPORT FIO;

CONST
  in  = FIO.StandardInput;

  debug = FALSE;

VAR
  done: BOOLEAN;


PROCEDURE WrRegs( f: FIO.File; ip, sp, bp: TAddress; ax: INTEGER );
BEGIN
  FIO.WrStr( f, '@ IP ' ); FIO.WrCard( f, ip, 4 );
  FIO.WrStr( f, '  SP ' ); FIO.WrCard( f, sp, 4 );
  FIO.WrStr( f, '  BP ' ); FIO.WrCard( f, bp, 4 );
  FIO.PrefixChar := ' ';
  FIO.WrStr( f, '  AX ' ); FIO.WrInt( f, ax, 4 );
  FIO.WrStr( f, '  ' );
END WrRegs;

PROCEDURE WrLn( f: FIO.File );
BEGIN
  FIO.WrLn( f );
END WrLn;

PROCEDURE WrIns( f: FIO.File; ip: TAddress );
VAR ins: TInstruction;
BEGIN
  FIO.WrCard( f, ip, 4 );
  FIO.WrStr( f, ':  ' );
  ins := code[ ip ];
  CASE ins.fct OF
  | fLit: FIO.WrStr( f, 'Put   ' ); FIO.WrInt( f, ins.val, 0 );
  | fOpr: CASE ins.opr OF
          | opRet: FIO.WrStr( f, 'Leave' );
          | opNeg: FIO.WrStr( f, 'Neg' );
          | opAdd: FIO.WrStr( f, 'Add' );
          | opSub: FIO.WrStr( f, 'Sub' );
          | opMul: FIO.WrStr( f, 'Mul' );
          | opDiv: FIO.WrStr( f, 'Div' );
          | opOdd: FIO.WrStr( f, 'Odd' );
          | opNop: FIO.WrStr( f, 'Nop' );
          | opEql: FIO.WrStr( f, 'Eql' );
          | opNeq: FIO.WrStr( f, 'Neq' );
          | opLss: FIO.WrStr( f, 'Lss' );
          | opGeq: FIO.WrStr( f, 'Geq' );
          | opGtr: FIO.WrStr( f, 'Gtr' );
          | opLeq: FIO.WrStr( f, 'Leq' );
          | opRed: FIO.WrStr( f, 'Read' );
          | opWrt: FIO.WrStr( f, 'Write' );
          ELSE
            FIO.WrStr( f, 'invalid opeartion' );
          END;
  | fLod: FIO.WrStr( f, 'Load  ' ); FIO.WrCard( f, ins.lvl, 0 );
          FIO.WrStr( f, ', ' ); FIO.WrCard( f, ins.adr, 0 );
  | fSto: FIO.WrStr( f, 'Store ' ); FIO.WrCard( f, ins.lvl, 0 );
          FIO.WrStr( f, ', ' ); FIO.WrCard( f, ins.adr, 0 );
  | fCal: FIO.WrStr( f, 'Call  ' ); FIO.WrCard( f, ins.lvl, 0 );
          FIO.WrStr( f, ', ' ); FIO.WrCard( f, ins.adr, 4 );
  | fInt: FIO.WrStr( f, 'Enter ' ); FIO.WrCard( f, ins.val, 0 );
  | fJmp: FIO.WrStr( f, 'Jmp   ' ); FIO.WrCard( f, ins.adr, 4 );
  | fJpc: FIO.WrStr( f, 'JmpZ  ' ); FIO.WrCard( f, ins.adr, 4 );
  | fStp: FIO.WrStr( f, 'Stop' );
  ELSE
    FIO.WrStr( f, 'invalid instruction' );
  END;
  WrLn( f );
END WrIns;


PROCEDURE ReadInt( VAR x: INTEGER );
BEGIN
  FIO.WrStr( out, '> ' );
  x := FIO.RdInt( in );
  done := FIO.OK;
END ReadInt;

PROCEDURE WriteInt( x, l: INTEGER );
BEGIN
  FIO.WrInt( out, x, l );
  FIO.WrLn( out );
END WriteInt;

PROCEDURE List( f: FIO.File );
VAR
  ip: CARDINAL;
  fct: TFunction;
BEGIN
  FIO.WrStr( f, '# Listing:' );
  FIO.WrLn( f );
  ip := MIN( TAddress );
  REPEAT
    fct := code[ ip ].fct;
    WrIns( f, ip );
    INC( ip );
  UNTIL ( fct >= fStp ) OR ( ip > MAX(TAddress) );
  FIO.WrStr( f, '# Listing of ' );
  FIO.WrCard( f, ip, 0 );
  FIO.WrStr( f, ' instructions done.' );
  FIO.WrLn( f );
END List;

PROCEDURE Run();
VAR
  ip,cnt: CARDINAL;
  bp, sp: TAddress;
  ins: TInstruction;
  stk: ARRAY TAddress OF INTEGER;
  stop: BOOLEAN;

  PROCEDURE Base( l: TLevel ): TAddress;
  VAR p: TAddress;
  BEGIN
    p := bp;
    WHILE l > 0 DO
      p := stk[ p ]; DEC( l );
    END;
    RETURN p
  END Base;

BEGIN
  FIO.WrStr( out, '# Interpreting:' );
  FIO.WrLn( out );
  ip := 0;
  sp := 0;
  bp := 1;
  cnt := 0;
  stk[ 1 ] := 0;
  stk[ 2 ] := 0;
  stk[ 3 ] := 0;
  stop := FALSE;
  REPEAT
    IF debug THEN WrRegs( dbg, ip, sp, bp, stk[sp] ); WrIns( dbg, ip ) END;
    ins := code[ip]; INC(ip);
    CASE ins.fct OF
    | fLit: INC(sp); stk[sp] := ins.val;
    | fOpr: CASE ins.opr OF
            | opRet: sp := bp-1; ip := stk[sp+3]; bp := stk[sp+2];
            | opNeg: stk[sp] := -stk[sp];
            | opAdd: DEC(sp); stk[sp] := stk[sp]  +  stk[sp+1];
            | opSub: DEC(sp); stk[sp] := stk[sp]  -  stk[sp+1];
            | opMul: DEC(sp); stk[sp] := stk[sp]  *  stk[sp+1];
            | opDiv: DEC(sp); stk[sp] := stk[sp] DIV stk[sp+1];
            | opOdd: stk[sp] := ORD( ODD(stk[sp]) );
            | opNop: ;
            | opEql: DEC(sp); stk[sp] := ORD( stk[sp] =  stk[sp+1] );
            | opNeq: DEC(sp); stk[sp] := ORD( stk[sp] #  stk[sp+1] );
            | opLss: DEC(sp); stk[sp] := ORD( stk[sp] <  stk[sp+1] );
            | opGeq: DEC(sp); stk[sp] := ORD( stk[sp] >= stk[sp+1] );
            | opGtr: DEC(sp); stk[sp] := ORD( stk[sp] >  stk[sp+1] );
            | opLeq: DEC(sp); stk[sp] := ORD( stk[sp] <= stk[sp+1] );
            | opRed: INC(sp); ReadInt( stk[sp] ); stop := ~done;
            | opWrt: WriteInt( stk[sp], 7 ); DEC(sp);
            END;
    | fLod: INC(sp); stk[sp] := stk[ Base(ins.lvl) + ins.adr ];
    | fSto: stk[ Base(ins.lvl) + ins.adr ] := stk[sp]; DEC(sp);
    | fCal: stk[sp+1] := Base(ins.lvl); stk[sp+2] := bp; stk[sp+3] := ip;
            bp := sp+1; ip := ins.adr; (* sp set with local vars *)
    | fInt: sp := sp + ins.adr;
    | fJmp: ip := ins.adr;
    | fJpc: IF stk[sp] = 0 THEN ip:= ins.adr END; DEC(sp);
    | fStp: stop := TRUE;
    END;
    INC( cnt );
  UNTIL stop OR ( ip > MAX(TAddress) );
  FIO.WrStr( out, '# Interpretation of ' );
  FIO.WrCard( out, cnt, 0 );
  FIO.WrStr( out, ' instructions done.' );
  FIO.WrLn( out );
END Run;

BEGIN
  done := FALSE;
END PL0INT.
