IMPLEMENTATION MODULE PL0PAR;

FROM Storage IMPORT ALLOCATE, DEALLOCATE;
FROM PL0SCA IMPORT TSymbol, sym, id, num;
FROM PL0SCA IMPORT Diff, KeepId, Mark, GetSym;
FROM PL0SCA IMPORT out, err, WrId;
FROM PL0GEN IMPORT Label, Gen, Fixup;
FROM PL0GEN IMPORT Enter, Leave, Jmp, JmpZ, JmpFwd, JmpZFwd, Call, Stop;
FROM PL0GEN IMPORT Put, Read, Write, Load, Store;
FROM PL0GEN IMPORT Neg, Odd, Add, Sub, Mul, Div;
FROM PL0GEN IMPORT Eql, Neq, Lss, Leq, Gtr, Geq;
FROM PL0INT IMPORT TAddress;

TYPE
  TObjClass = ( ocConst, ocVar, ocProc, ocHdr );
  TObj = POINTER TO TObjDesc;
  TObjDesc = RECORD
    name: CARDINAL;
    next: TObj;
    CASE kind: TObjClass OF
    | ocConst: val: INTEGER;
    | ocVar  : lvl, ofs: CARDINAL;
    | ocProc : plvl, adr, size: CARDINAL;
    | ocHdr  : last, down: TObj;
    END;
  END;

VAR
  topScope, undef: TObj;
  lvl: CARDINAL;


PROCEDURE DoFind( o: TObj ): TObj;
BEGIN
  WHILE o # NIL DO
    IF Diff( id, o^.name ) = 0 THEN
      RETURN o;
    ELSE
      o := o^.next;
    END;
  END;
  RETURN o;
END DoFind;

PROCEDURE Err( n: CARDINAL );
BEGIN
  noErr := FALSE;
  Mark( n );
END Err;

PROCEDURE Test( s: TSymbol; n: CARDINAL );
BEGIN
  IF sym < s THEN
    Err( n ); REPEAT GetSym() UNTIL sym >= s;
  END;
END Test;

PROCEDURE NewObj( k: TObjClass ): TObj;
VAR o: TObj;
BEGIN
  o := DoFind( topScope^.next );
  IF o # NIL THEN Err( 25 ) END;
  ALLOCATE( o, SIZE(o^) );
  WITH o^ DO name := id; kind := k; next := NIL END;
  KeepId();
  topScope^.last^.next := o;
  topScope^.last := o;
  RETURN o;
END NewObj;

PROCEDURE Find( id: CARDINAL ): TObj;
VAR h, o: TObj;
BEGIN
  h := topScope;
  WHILE h # NIL DO
    o := DoFind( h^.next );
    IF o # NIL THEN RETURN o END;
    h := h^.down;
  END;
  Err( 11 );
  RETURN undef;
END Find;

PROCEDURE Expression();

  PROCEDURE Factor();
  VAR o: TObj;
  BEGIN
    Test( lparen, 6 );
    IF sym = ident THEN
      o := Find( id );
      IF o # NIL THEN
        CASE o^.kind OF
        | ocConst: Put( o^.val );
        | ocVar  : Load( lvl - o^.lvl, o^.ofs );
        | ocProc : Err( 21 );
        END;
      END;
      GetSym();
    ELSIF sym = number THEN
      Put( num ); GetSym();
    ELSIF sym = lparen THEN
      GetSym(); Expression();
      IF sym = rparen THEN GetSym() ELSE Err( 7 ) END;
    ELSE
      Err( 8 );
    END;
  END Factor;

  PROCEDURE Term();
  VAR op: TSymbol;
  BEGIN
    Factor();
    WHILE ( times <= sym ) & ( sym <= div ) DO  (* some mul op *)
      op := sym; GetSym(); Factor();
      IF op = times THEN Mul() ELSE Div() END;
    END;
  END Term;

VAR op: TSymbol;
BEGIN
  IF ( plus <= sym ) & ( sym <= minus ) THEN  (* some sign op *)
    op := sym; GetSym();
    Term();
    IF op = minus THEN Neg() END;
  ELSE
    Term();
  END;
  WHILE ( plus <= sym ) & ( sym <= minus ) DO  (* some add op *)
    op := sym; GetSym(); Term();
    IF op = plus THEN Add() ELSE Sub() END;
  END;
END Expression;

PROCEDURE Condition();
VAR op: TSymbol;
BEGIN
  IF sym = odd THEN
    GetSym(); Expression(); Odd();
  ELSE
    Expression();
    IF ( eql <= sym ) & ( sym <= geq ) THEN  (* some cmp op *)
      op := sym; GetSym(); Expression();
      CASE op OF
      | eql: Eql();
      | neq: Neq();
      | lss: Lss();
      | geq: Geq();
      | gtr: Gtr();
      | leq: Leq();
      ELSE
        HALT;
      END;
    ELSE
      Err( 20 );
    END;
  END;
END Condition;

PROCEDURE Statement();
VAR
  o: TObj;
  entry, from: TAddress;
BEGIN
  IF sym = ident THEN
    o := Find( id );
    IF o^.kind # ocVar THEN Err( 12 ); o := NIL END;
    GetSym();
    IF sym = becomes THEN
      GetSym();
    ELSIF sym = eql THEN
      Err( 13 ); GetSym();
    ELSE
      Err( 13 );
    END;
    Expression();
    IF o # NIL THEN Store( lvl - o^.lvl, o^.ofs ) END;
  ELSIF sym = call THEN
    GetSym();
    IF sym = ident THEN
      o := Find( id );
      IF o^.kind = ocProc THEN
        Call( lvl - o^.lvl, o^.adr );
      ELSE
        Err( 15 );
      END;
      GetSym();
    ELSE
      Err( 14 );
    END;
  ELSIF sym = begin THEN
    GetSym();
    LOOP
      Statement();
      IF sym = semicolon THEN
        GetSym();
      ELSIF sym = end THEN
        GetSym(); EXIT;
      ELSIF sym < const THEN
        Err( 17 );
      ELSE
        Err( 17 ); EXIT;
      END;
    END;
  ELSIF sym = if THEN
    GetSym(); Condition();
    IF sym = then THEN GetSym() ELSE Err( 16 ) END;
    JmpZFwd( from ); Statement(); Fixup( from );
  ELSIF sym = while THEN
    entry := Label(); GetSym(); Condition();
    IF sym = do THEN GetSym() ELSE Err( 18 ) END;
    JmpZFwd( from ); Statement(); Jmp( entry ); Fixup( from );
  ELSIF sym = read THEN
    GetSym();
    IF sym = ident THEN
      o := Find( id );
      IF o^.kind = ocVar THEN
        Read(); Store( lvl - o^.lvl, o^.ofs );
      ELSE
        Err( 12 );
      END;
    ELSE
      Err( 14 );
    END;
    GetSym();
  ELSIF sym = write THEN
    GetSym(); Expression(); Write();
  END;
  Test( ident, 19 );
END Statement;

PROCEDURE Block();
VAR
  ofs  : CARDINAL;  (* stack offset *)
  entry: TAddress;

  PROCEDURE ConstDeclaration();
  VAR o: TObj;
  BEGIN
    IF sym = ident THEN
      GetSym();
      IF ( sym = eql ) OR ( sym = becomes ) THEN
        IF sym = becomes THEN Err( 1 ) END;
        GetSym();
        IF sym = number THEN
          o := NewObj( ocConst ); o^.val := num;
          GetSym();
        ELSE
          Err( 2 );
        END;
      ELSE
        Err( 3 );
      END;
    ELSE
      Err( 4 );
    END;
  END ConstDeclaration;

  PROCEDURE VarDeclaration();
  VAR o: TObj;
  BEGIN
    IF sym = ident THEN
      o := NewObj( ocVar );
      o^.lvl := lvl; o^.ofs := ofs; INC( ofs );
      GetSym();
    ELSE
      Err( 4 );
    END;
  END VarDeclaration;

VAR h,o: TObj;
BEGIN
  INC( lvl );
  ofs := 3;                      (* stack frame uses 3 units *)
  ALLOCATE( h, SIZE(h^) );
  WITH h^ DO
    kind := ocHdr; next := NIL; last := h; name := 0; down := topScope;
  END;
  topScope := h;
  JmpFwd( entry );
  IF sym = const THEN
    GetSym();
    LOOP
      ConstDeclaration();
      CASE sym OF
      | comma    : GetSym();
      | semicolon: GetSym(); EXIT;
      | ident    : Err( 5 );
      ELSE
        Err( 5 ); EXIT;
      END;
    END;
  END;
  IF sym = var THEN
    GetSym();
    LOOP
      VarDeclaration();
      CASE sym OF
      | comma    : GetSym();
      | semicolon: GetSym(); EXIT;
      | ident    : Err( 5 );
      ELSE
        Err( 5 ); EXIT;
      END;
    END;
  END;
  WHILE sym = procedure DO
    GetSym();
    IF sym = ident THEN GetSym() ELSE Err( 4 ) END;
    o := NewObj( ocProc );
    o^.plvl := lvl; o^.adr := Label();  (* nested code comes next *)
    IF sym = semicolon THEN GetSym() ELSE Err( 5 ) END;
    Block();
    IF sym = semicolon THEN GetSym() ELSE Err( 5 ) END;
  END;
  Fixup( entry );
  Enter( ofs );
  Statement();
  IF lvl > 1 THEN Leave() ELSE Stop() END;
  topScope := topScope^.down;
  DEC( lvl );
END Block;

PROCEDURE Run();
BEGIN
  noErr := TRUE;
  topScope := NIL;
  lvl := 0;
  GetSym();
  Block();
  IF sym # period THEN Err( 9 ) END;
END Run;

PROCEDURE End();
BEGIN
END End;

BEGIN
  ALLOCATE( undef, SIZE(undef^) );
  WITH undef^ DO
    name := 0; next := NIL; kind := ocVar; lvl := 0; ofs := 0;
  END;
END PL0PAR.
