IMPLEMENTATION MODULE PL0SCA;

FROM SYSTEM IMPORT ADR;
IMPORT Str, Lib, FIO;

CONST
  maxCard = 0FFFFH;  (* MAX(CARDINAL) would allow for 32-bit-data *)
  bufLen = 1000;
  maxKwCnt = 20;  (* structured constants, how I miss you! *)
  chEOF = 32C;    (* = 1AH = 26 = ASCII.sub (TS) *)
  (* TopSpeed's FileSystem.ReadChar() returns CHR(26) while Wirth used 0C  *)
  chLF = 12C;

TYPE
  TKwDesc = RECORD
    sym: TSymbol;
    id : CARDINAL;
  END;
  TKwIdx = [ 1 .. maxKwCnt ];
  TKwTab = RECORD
    cnt: CARDINAL;
    arr: ARRAY TKwIdx OF TKwDesc;
  END;

  TBufIdx = [ 0 .. bufLen - 1 ];
  TBufDesc = ARRAY TBufIdx OF CHAR;
  (* layout: | KW KW .. KW | ID .. ID | EMPTY | *)
  (*                         ^id0       ^id1    *)

VAR
  ch: CHAR;
  id0, id1: CARDINAL;
  kwTab: TKwTab;
  buf: TBufDesc;

PROCEDURE WrId( f: FIO.File );
VAR i,c: CARDINAL;
BEGIN
  i := id;
  c := ORD( buf[id] );
  WHILE c > 1 DO
    INC( i );
    FIO.WrChar( f, buf[i] );
    DEC( c );
  END;
END WrId;

PROCEDURE Fatal( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'PL0SCA.Fatal: ' );
  FIO.WrStr( err, s );
  FIO.WrLn( err );
  HALT;
END Fatal;

PROCEDURE Mark( n: CARDINAL );
BEGIN
  FIO.WrStr( out, '[' );
  FIO.WrCard( out, n, 0 );
  FIO.WrStr( out, ']' );
END Mark;

PROCEDURE GetCh();
BEGIN
  ch := FIO.RdChar( src );
  IF ( ' ' <= ch ) & ( ch < 177C ) THEN
    FIO.WrChar( out, ch );
  ELSIF ch = chLF THEN
    FIO.WrLn( out );
  END;
END GetCh;

PROCEDURE Diff( u,v: CARDINAL ): INTEGER;
VAR w: CARDINAL;
BEGIN
  (* called from Identifier() only, so u,v both represent valid strings *)
  w := ORD( buf[u] );   (* w # 0, same reason *)
  LOOP
    IF w = 0 THEN
      RETURN 0;
    ELSIF buf[ u ] # buf[ v ] THEN
      RETURN INTEGER( ORD(buf[u]) ) - INTEGER( ORD(buf[v]) );
    ELSE
      INC( u ); INC( v ); DEC( w );
    END;
  END;
END Diff;

PROCEDURE KeepId();
BEGIN
  id := id1;
END KeepId;

PROCEDURE NotAlpha(): BOOLEAN;
BEGIN
  RETURN ( ch < '0' ) OR ( '9' < ch ) & ( CAP(ch) < 'A' ) OR ( 'Z' < CAP(ch) );
END NotAlpha;

PROCEDURE Identifier();
VAR a, b, m: CARDINAL; d: INTEGER;
BEGIN
  sym := ident;
  id1 := id;
  INC( id1 );
  REPEAT
    IF id1 > MAX(TBufIdx) THEN Fatal( 'buffer full' ) END;
    buf[ id1 ] := ch; INC( id1 );
    GetCh();
  UNTIL NotAlpha();
  buf[ id ] := CHR( id1 - id );  (* entry size (incl. length byte) *)
  a := 1; b := kwTab.cnt;
  REPEAT
    m := ( a + b ) DIV 2;
    d := Diff( id, kwTab.arr[m].id );
    IF d <= 0 THEN b := m - 1 END;
    IF d >= 0 THEN a := m + 1 END;
  UNTIL a > b;
  IF a > b + 1 THEN
    sym := kwTab.arr[ m ].sym
  END;
END Identifier;

PROCEDURE Number();
VAR
  d: CARDINAL;
BEGIN
  sym := number;
  num := 0;
  REPEAT
    d := ORD( ch ) - ORD( '0' );
    IF ( maxCard - d ) DIV 10 >= num THEN
      num := 10 * num + d;
    ELSE
      Mark( 30 );
      num := 0;
    END;
    GetCh();
  UNTIL NotAlpha();
END Number;

PROCEDURE GetSym();

  PROCEDURE Comment();
  BEGIN
    GetCh();
    REPEAT
      WHILE ch # '*' DO GetCh() END;
      GetCh();
    UNTIL ch = ')';
    GetCh()
  END Comment;

BEGIN
  LOOP (* ignore control characters *)
    IF ch <= ' ' THEN
      IF ch = chEOF THEN ch := ' '; EXIT END;
      GetCh();
    ELSIF ch >= 177C THEN
      GetCh();
    ELSE
      EXIT;
    END
  END;
  CASE ch OF   (* " " <= ch < 177C *)
  | ' '     : sym := eof;    ch := chEOF;
  | '!'     : sym := write;  GetCh();
  | '"'     : sym := null;   GetCh();
  | '#'     : sym := neq;    GetCh();
  | '$'.."'": sym := null;   GetCh();
  | '('     : GetCh();
              IF ch = '*' THEN Comment(); GetSym() ELSE sym := lparen END;
  | ')'     : sym := rparen; GetCh();
  | '*'     : sym := times;  GetCh();
  | '+'     : sym := plus;   GetCh();
  | ','     : sym := comma;  GetCh();
  | '-'     : sym := minus;  GetCh();
  | '.'     : sym := period; GetCh();
  | '/'     : sym := div;    GetCh();
  | '0'..'9': Number();
  | ':'     : GetCh();
              IF ch = '=' THEN GetCh(); sym := becomes ELSE sym := null END;
  | ';'     : sym := semicolon; GetCh();
  | '<'     : GetCh();
              IF ch = '=' THEN GetCh(); sym := leq ELSE sym := lss END;
  | '='     : sym := eql; GetCh();
  | '>'     : GetCh();
              IF ch = '=' THEN GetCh(); sym := geq ELSE sym := gtr END;
  | '?'     : sym := read; GetCh();
  | '@'     : sym := null; GetCh();
  | 'A'..'Z': Identifier();
  | '['..'`': sym := null; GetCh();
  | 'a'..'z': Identifier();
  | '{'..'~': sym := null; GetCh();
  END;
END GetSym;

PROCEDURE Init();
BEGIN
  id := id0;
  ch := ' ';
END Init;

PROCEDURE Close();
BEGIN
  FIO.WrLn( out );
END Close;

PROCEDURE EnterKW( sym: TSymbol; str: ARRAY OF CHAR );
VAR i, j: CARDINAL;
BEGIN
  INC( kwTab.cnt );
  IF kwTab.cnt > MAX( TKwIdx ) THEN Fatal( 'debug TKwIdx' ) END;
  kwTab.arr[ kwTab.cnt ].sym := sym;
  kwTab.arr[ kwTab.cnt ].id  := id;
  i := 0; j := Str.Length( str ); (* mocka's HIGH() is buggy *)
  buf[ id ] := CHR( j + 1 );    (* entry size (incl. length byte) *)
  INC( id );
  WHILE i < j DO
    buf[ id ] := str[ i ];
    INC( id ); INC( i );
  END;
  id0 := id;
END EnterKW;

BEGIN
  sym := null;
  id := 0;
  num := 0;
  src := FIO.StandardInput;
  ch := ' ';
  id0 := 0; id1 := 0;
  Lib.Fill( ADR(kwTab), SIZE(kwTab), 0C );
  Lib.Fill( ADR(buf), SIZE(buf), 0C );
  EnterKW( do, 'DO' );
  EnterKW( if, 'IF' );
  EnterKW( end, 'END' );
  EnterKW( odd, 'ODD' );
  EnterKW( var, 'VAR' );
  EnterKW( call, 'CALL' );
  EnterKW( then, 'THEN' );
  EnterKW( begin, 'BEGIN' );
  EnterKW( const, 'CONST' );
  EnterKW( while, 'WHILE' );
  EnterKW( procedure, 'PROCEDURE' );
END PL0SCA.
